@@ -28,7 +28,15 @@ angular.module('index').controller('indexController', ['$scope', '$injector',
     var $window          = $injector.get('$window');
     var clipboardService = $injector.get('clipboardService');
     var guacNotification = $injector.get('guacNotification');
-    
+
+    /**
+     * The error that prevents the current page from rendering at all. If no
+     * such error has occurred, this will be null.
+     *
+     * @type Error
+     */
+    $scope.fatalError = null;
+
     /**
      * The notification service.
      */
@@ -130,6 +138,12 @@ angular.module('index').controller('indexController', ['$scope', '$injector',
         keyboard.reset();
     };
 
+    // Release all keys upon form submission (there may not be corresponding
+    // keyup events for key presses involved in submitting a form)
+    $document.on('submit', function formSubmitted() {
+        keyboard.reset();
+    });
+
     /**
      * Checks whether the clipboard data has changed, firing a new
      * "guacClipboard" event if it has.
@@ -159,6 +173,7 @@ angular.module('index').controller('indexController', ['$scope', '$injector',
         $scope.loginHelpText = null;
         $scope.acceptedCredentials = {};
         $scope.expectedCredentials = error.expected;
+        $scope.fatalError = null;
     });
 
     // Prompt for remaining credentials if provided credentials were not enough
@@ -168,6 +183,20 @@ angular.module('index').controller('indexController', ['$scope', '$injector',
         $scope.loginHelpText = error.translatableMessage;
         $scope.acceptedCredentials = parameters;
         $scope.expectedCredentials = error.expected;
+        $scope.fatalError = null;
+    });
+
+    // Replace absolutely all content with an error message if the page itself
+    // cannot be displayed due to an error
+    $scope.$on('guacFatalPageError', function fatalPageError(error) {
+        $scope.page.title = 'APP.NAME';
+        $scope.page.bodyClassName = '';
+        $scope.fatalError = error;
+    });
+
+    // Ensure new pages always start with clear keyboard state
+    $scope.$on('$routeChangeStart', function routeChanging() {
+        keyboard.reset();
     });
 
     // Update title and CSS class upon navigation
@@ -181,6 +210,7 @@ angular.module('index').controller('indexController', ['$scope', '$injector',
             $scope.loginHelpText = null;
             $scope.acceptedCredentials = null;
             $scope.expectedCredentials = null;
+            $scope.fatalError = null;
 
             // Set title
             var title = current.$$route.title;
